import os
import pickle
import time
from pysimlink.utils import annotation_utils as anno
from pysimlink.lib.model_types import DataType
import numpy as np
import warnings


def infer_defines(model_paths: "anno.ModelPaths"):
    """When defines.txt is not present, add the only required defines for pysimlink

    Args:
        model_paths: instance of ModelPaths object pointing to the root _model

    Returns:
        List of _model defines. These are added the CMakeLists.txt

    """
    ret = [f"MODEL={model_paths.root_model_name}"]
    return ret


def print_all_params(model: "anno.Model"):
    """
    Prints all parameters for the given model.

    Uses the ModelInfo object to print all model info about the root and each reference model

    Args:
        model: instance of the Model to print params of
    """
    params = model.get_params()
    for model_info in params:
        print(f"Parameters for model at '{model_info.model_name}'")
        print("  model parameters:")
        for param in model_info.model_params:
            print(f"    param: '{param.model_param}' | data_type: '{DataType(param.data_type)}'")
        print("  block parameters:")
        for param in model_info.block_params:
            print(
                f"    Block: '{param.block_name}' | Parameter: '{param.block_param}' | data_type: '{DataType(param.data_type)}'"
            )
        print("  signals:")
        for sig in model_info.signals:
            print(
                f"    Block: '{sig.block_name}' | Signal Name: '{sig.signal_name}' | data_type: '{DataType(sig.data_type)}'"
            )
        print("-" * 80)


def get_other_in_dir(directory: str, known: str):
    """In a directory containing only two directories, get the name of the other we don't know

    Args:
        directory: path to the directory
        known: The file/folder known to exist in the directory

    Returns:
         the other directory/file in the directory
    """

    model_folders = set(os.listdir(directory))
    model_folders.discard(".DS_Store")
    assert (
        len(model_folders) == 2
    ), f"Directory '{directory}' contains more than 2 folders (not counting .DS_Store on Mac)"
    assert (
        known in model_folders
    ), f"File does not exist in {directory}. Should be one of {model_folders}"
    model_folders.remove(known)

    return model_folders.pop()


def with_read_lock(func: callable) -> callable:
    """Use as decorator (@with_lock) around object methods that need locking.

    Note: The object must have a self._lock property.
    Locking thus works on the object level (no two locked methods of the same
    object can be called asynchronously).

    Inspired by `Rllib <https://github.com/ray-project/ray/blob/4963dfaae0fbdbae4a5ad6188bc86986f1a9568a/rllib/utils/threading.py#L7>`_

    Args:
        func: The function to decorate/wrap.
    Returns:
        The wrapped (object-level locked) function.
    """

    def wrapper(self, *a, **k):
        try:
            with self._lock.read_lock():
                return func(self, *a, **k)
        except AttributeError as e:
            if "has no attribute '_lock'" in e.args[0]:
                raise AttributeError(
                    "Object {} must have a `self._lock` property (assigned "
                    "to a fasteners.InterProcessReaderWriterLock object in its "
                    "constructor)!".format(self)
                )
            raise e

    return wrapper


def mt_rebuild_check(model_paths: "anno.ModelPaths", force_rebuild: bool) -> bool:
    """
    Prevent the model from being rebuilt in every multithreading instance

    Args:
        model_paths (anno.ModelPaths): instance of the model paths object. Used to get the tmp_dir
        force_rebuild (bool): flag set by the user that forces the model to rebuild

    Returns:
         True if the model should rebuild because of the force_rebuild flag and has not already
    """
    if not force_rebuild:
        return False

    compile_info = os.path.join(model_paths.tmp_dir, "compile_info.pkl")
    if not os.path.exists(compile_info):
        return True

    with open(compile_info, "rb") as f:
        info = pickle.load(f)

    if info["parent"] == os.getppid():
        tdiff = time.time() - info["time"]

        # assume that it takes at least 1 second to start a separate instance of a program
        # If it takes less than 1 second, then we assume it is run within the same python
        # instance
        return tdiff > 1.0
    else:
        return True


def sanitize_model_name(model_name):
    return model_name.replace(" ", "").replace("-", "_").lower()


def cast_type(value: "anno.Value", dtype: "anno.DataType", orientations):
    if not isinstance(value, np.ndarray):
        value_cp = np.array(value, dtype=dtype.pythonType)
        if value_cp != value:
            warnings.warn(
                f"Datatype of value does not match parameter: Data loss detected! ({value} -> {value_cp})",
                RuntimeWarning,
                stacklevel=3,
            )
        value = value_cp

    if str(value.dtype) != dtype.pythonType:
        warnings.warn(
            f"Datatype of value does not match parameter. Expected {dtype.pythonType} got {value.dtype}",
            RuntimeWarning,
            stacklevel=3,
        )

    if dtype.dims != value.shape:
        # We can perform reshaping and formatting in one operation
        orientation = (
            "F" if dtype.orientation in [orientations.col_major_nd, orientations.col_major] else "C"
        )
        value = np.reshape(value, tuple(dtype.dims), orientation)
        if str(value.dtype) != dtype.pythonType:
            value = value.astype(dtype.pythonType)
    else:
        ## We can perform type casting and formatting in one operation
        if dtype.orientation in [orientations.col_major_nd, orientations.col_major]:
            value = np.asfortranarray(value, dtype=dtype.pythonType)
        elif dtype.orientation in [orientations.row_major_nd, orientations.row_major]:
            value = np.ascontiguousarray(value, dtype=dtype.pythonType)
        elif str(value.dtype) != dtype.pythonType:
            value = value.astype(dtype.pythonType)

    return value



import glob
import os

from pysimlink.lib import cmake_gen
from pysimlink.lib.compilers.compiler import Compiler
from pysimlink.utils import annotation_utils as anno


class NoRefCompiler(Compiler):
    """
    Compiler for a model that does not use model references
    """

    def __init__(self, model_paths: "anno.ModelPaths", generator: str):
        super().__init__(model_paths, generator)
        self.model_srcs = []
        self.model_incs = []

    def compile(self):
        self.clean()
        self._get_simulink_deps()
        self._gen_custom_srcs()
        self._gen_model_deps()
        self._gen_cmake()
        self._build()

    def _gen_model_deps(self):
        self.model_srcs = glob.glob(self.model_paths.root_model_path + "/**/*.c", recursive=True)
        self.model_incs = []
        for dir_name in os.walk(self.model_paths.root_model_path, followlinks=False):
            for file in dir_name[2]:
                if ".h" in file:
                    self.model_incs.append(dir_name[0])
                    break

    def _gen_cmake(self):
        includes = [self.custom_includes] + self.model_incs
        for dir_name in os.walk(self.model_paths.simulink_native, followlinks=False):
            for file in dir_name[2]:
                if ".h" in file:
                    includes.append(dir_name[0])
                    break
        maker = cmake_gen.CmakeTemplate(
            self.model_paths.root_model_name.replace(" ", "_").replace("-", "_").lower()
        )
        cmake_text = maker.header()
        cmake_text += maker.set_includes(includes)

        cmake_text += maker.add_library(self.model_paths.root_model_name, self.model_srcs)
        cmake_text += maker.add_library("shared_utils", self.simulink_deps_src)
        cmake_text += maker.add_custom_libs(self.custom_sources)
        cmake_text += maker.set_lib_props()

        dep_map = {self.model_paths.root_model_name: ["shared_utils"]}
        cmake_text += maker.add_link_libs(dep_map)
        cmake_text += maker.add_private_link(self.model_paths.root_model_name)
        cmake_text += maker.set_lib_props()
        cmake_text += maker.add_compile_defs(self.defines)

        cmake_text += maker.footer()

        with open(
            os.path.join(self.model_paths.tmp_dir, "CMakeLists.txt"), "w", encoding="utf-8"
        ) as f:
            f.write(cmake_text)

    def gather_types(self):
        types_files = glob.glob(self.model_paths.root_model_path + "/*_types.h")

        for file in types_files:
            with open(file, "r") as f:
                lines = f.readlines()

            self._read_types_single_file(lines)

        return self._gen_types()



import os

class GenerationError(Exception):
    """
    Exception raised when generating build files fails.

    Attributes:
        dump (str): stdout and stderr output of the **generation** process
        cmake (str): contents of the :file:`CMakeLists.txt` file
    """

    def __init__(self, *args):
        self.dump = args[0]
        self.cmake = args[1]

    def __str__(self):
        if os.environ.get("PYSIMLINK_DEBUG", "FALSE") == "TRUE":
            with open(self.cmake, 'r') as f:
                contents = f.read()
            with open(self.dump, 'r') as f:
                dump = f.read()
            ret = f"GENERATION ERROR\n\n{dump}\nCMAKE_LISTS\n{contents}"
            return ret
        else:
            return (
                "Generating the CMakeLists for this model failed. This could be a c/c++/cmake setup issue, bad paths, or a bug! "
                f"Output from CMake generation is in {self.dump}"
            )


class BuildError(Exception):
    """
    Exception raised when generating build files succeeded but compiling the model fails.

    Attributes:
        dump (str): stdout and stderr output of the **build** process
        cmake (str): contents of the :file:`CMakeLists.txt` file
    """

    def __init__(self, *args):
        self.dump = args[0]
        self.cmake = args[1]

    def __str__(self):
        if os.environ.get("PYSIMLINK_DEBUG", "FALSE") == "TRUE":
            with open(self.cmake, 'r') as f:
                contents = f.read()
            with open(self.dump, 'r') as f:
                dump = f.read()
            ret = f"BUILD ERROR\n\n{dump}\n\n---\nCMAKE_LISTS\n{contents}"
            return ret
        else:
            return (
                "Building the model failed. This could be a c/c++/cmake setup issue, bad paths, or a bug! "
                f"Output from the build process is in {self.dump}"
            )



from .lib.model import Model
from .lib.exceptions import BuildError, GenerationError
from .utils.model_utils import print_all_params
from .lib import model_types as types
from .utils import annotation_utils as anno

__all__ = ["Model", "BuildError", "GenerationError", "print_all_params", "types", "anno"]



import argparse

from pysimlink import Model, print_all_params


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("model_name")
    parser.add_argument("model_path")

    args = parser.parse_args()

    return args


def main():
    args = parse_args()
    md = Model(args.model_name, args.model_path)
    md.reset()
    print_all_params(md)


if __name__ == "__main__":
    main()



from dataclasses import dataclass


@dataclass
class Field:
    type: str
    name: str


class Struct:
    def __init__(self, name, fields):
        self.name = name
        self.fields = fields


def parse_struct(lines):
    fields = []
    for i, line in enumerate(lines):
        if i == 0:
            continue
        line = line.strip().split()
        if line[0] == "}":
            struct = Struct(line[1][:-1], fields)
            return struct
        else:
            fields.append(Field(line[0], line[1][:-1]))



## Coppied from
# https://github.com/pypa/pip/blob/7efc8149162d9ae2f5a60cb346979bc8a6e81700/src/pip/_internal/cli/spinners.py

# Copyright (c) 2008-present The pip developers (see AUTHORS.txt file)

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import contextlib
import itertools
import sys
import time
import os
from typing import IO, Generator, Optional
import threading


class SpinnerInterface:
    def spin(self) -> None:
        raise NotImplementedError()

    def finish(self, final_status: str) -> None:
        raise NotImplementedError()


class InteractiveSpinner(SpinnerInterface):
    def __init__(
        self,
        message: str,
        file: Optional[IO[str]] = None,
        spin_chars: str = "-\\|/",
        # Empirically, 8 updates/second looks nice
        min_update_interval_seconds: float = 0.125,
    ):
        self._message = message
        if file is None:
            file = sys.stdout
        self._file = file
        self._rate_limiter = RateLimiter(min_update_interval_seconds)
        self._finished = False

        self._spin_cycle = itertools.cycle(spin_chars)

        self._file.write(self._message + " ... ")
        self._width = 0

    def _write(self, status: str) -> None:
        assert not self._finished
        # Erase what we wrote before by backspacing to the beginning, writing
        # spaces to overwrite the old text, and then backspacing again
        backup = "\b" * self._width
        self._file.write(backup + " " * self._width + backup)
        # Now we have a blank slate to add our status
        self._file.write(status)
        self._width = len(status)
        self._file.flush()
        self._rate_limiter.reset()

    def spin(self) -> None:
        if self._finished:
            return
        if not self._rate_limiter.ready():
            return
        self._write(next(self._spin_cycle))

    def finish(self, final_status: str) -> None:
        if self._finished:
            return
        self._write(final_status)
        self._file.write("\n")
        self._file.flush()
        self._finished = True


# Used for dumb terminals, non-interactive installs (no tty), etc.
# We still print updates occasionally (once every 60 seconds by default) to
# act as a keep-alive for systems like Travis-CI that take lack-of-output as
# an indication that a task has frozen.
class NonInteractiveSpinner(SpinnerInterface):
    def __init__(self, message: str, min_update_interval_seconds: float = 60.0) -> None:
        self._message = message
        self._finished = False
        self._rate_limiter = RateLimiter(min_update_interval_seconds)
        self._update("started")

    def _update(self, status: str) -> None:
        assert not self._finished
        self._rate_limiter.reset()

    def spin(self) -> None:
        if self._finished:
            return
        if not self._rate_limiter.ready():
            return
        self._update("still running...")

    def finish(self, final_status: str) -> None:
        if self._finished:
            return
        self._update(f"finished with status '{final_status}'")
        self._finished = True


class RateLimiter:
    def __init__(self, min_update_interval_seconds: float) -> None:
        self._min_update_interval_seconds = min_update_interval_seconds
        self._last_update: float = 0

    def ready(self) -> bool:
        now = time.time()
        delta = now - self._last_update
        return delta >= self._min_update_interval_seconds

    def reset(self) -> None:
        self._last_update = time.time()


def spin(spinner, event):
    while event.isSet():
        spinner.spin()
        time.sleep(0.126)


@contextlib.contextmanager
def open_spinner(message: str) -> Generator[SpinnerInterface, None, None]:
    # Interactive spinner goes directly to sys.stdout rather than being routed
    # through the logging system, but it acts like it has level INFO,
    # i.e. it's only displayed if we're at level INFO or better.
    # Non-interactive spinner goes through the logging system, so it is always
    # in sync with logging configuration.
    if sys.stdout.isatty():
        spinner: SpinnerInterface = InteractiveSpinner(message)
    else:
        spinner = NonInteractiveSpinner(message)
    event = threading.Event()
    try:
        event.set()
        t = threading.Thread(target=spin, args=(spinner, event))
        t.start()
        with hidden_cursor(sys.stdout):
            yield
    except KeyboardInterrupt:
        event.clear()
        t.join()
        spinner.finish("canceled")
        raise
    except Exception:
        event.clear()
        t.join()
        spinner.finish("error")
        raise
    else:
        event.clear()
        t.join()
        spinner.finish("done")


HIDE_CURSOR = "\x1b[?25l"
SHOW_CURSOR = "\x1b[?25h"


@contextlib.contextmanager
def hidden_cursor(file: IO[str]) -> Generator[None, None, None]:
    # The Windows terminal does not support the hide/show cursor ANSI codes,
    # even via colorama. So don't even try.
    if os.name == "nt":
        yield
    # We don't want to clutter the output with control characters if we're
    # writing to a file, or if the user is running with --quiet.
    # See https://github.com/pypa/pip/issues/3418
    elif not file.isatty():
        yield
    else:
        file.write(HIDE_CURSOR)
        try:
            yield
        finally:
            file.write(SHOW_CURSOR)



from setuptools import find_packages, setup

with open("requirements.txt", "r") as f:
    reqs = f.read().splitlines()

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name="pysimlink",
    version="1.2.1",
    author="Landon Harris",
    author_email="lharri73@vols.utk.edu",
    description="Compile, run, and interact with Simulink models natively in Python!",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=reqs,
    extras_require={
        "dev": [
            "pylint",
            "black",
            "sphinx_rtd_theme",
            "sphinx",
            "tqdm",
            "sphinx-toolbox",
            "sphinx-hoverxref",
            "readthedocs-sphinx-search",
        ]
    },
    include_package_data=True,
    python_requires=">=3.6",
    keywords=["Simulink"],
    license="GPLv3",
    classifiers=[
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Code Generators",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
    ],
    url="https://github.com/lharri73/PySimlink",
    project_urls={
        "Documentation": "https://lharri73.github.io/PySimlink/",
        "Source": "https://github.com/lharri73/PySimlink",
    },
)



from abc import abstractmethod
import glob
import os
from datetime import datetime
from subprocess import Popen, PIPE
import shutil
import re
import warnings

import cmake

from pysimlink.utils import annotation_utils as anno
from pysimlink.utils.model_utils import infer_defines, sanitize_model_name
from pysimlink.lib.exceptions import GenerationError, BuildError
from pysimlink.lib.struct_parser import parse_struct


class Compiler:
    """
    Base class to discover sources and compile a model
    """

    simulink_deps: "set[str]"  ## All header files created in the simulink common directory
    simulink_deps_path: "list[str]"  ## Path to the header files created in the simulink common dir
    simulink_deps_src: "list[str]"  ## Path to all source files created in the simulink common dir
    model_paths: "anno.ModelPaths"  ## Instance of the ModelPaths containing information about the directory structure
    defines: "list[str]"  ## All defines that should be set during _model compilation
    custom_includes: str  ## Include files directory defined by this python module
    custom_sources: str  ## Source files directory defined by this python module
    types: "list[anno.Struct]"  ## Name of types used by signals (usually manifested as busses).
    matlogging: bool  ## Whether matfile logging is enabled

    def __init__(self, model_paths: "anno.ModelPaths", generator: str):
        self.model_paths = model_paths
        self.types = []
        self.matlogging = False
        self.generator=generator

    def clean(self):
        """
        Remove all files from the temporary directory
        """
        self.model_paths.clean()

    def compile(self):
        """
        Builds the cmake file, calls cmake, and builds the extension.
        """
        raise NotImplementedError

    def gather_types(self):
        """
        Gather all the types (structs) used in all models
        """
        raise NotImplementedError

    def get_type_names(self):
        """
        return name of all structs used as bus names
        """
        raise NotImplementedError

    def needs_to_compile(self) -> bool:
        """
        check if the model extension exists.

        Returns:
            bool: True if the model needs to be compiled, False otherwise
        """
        if os.name == "nt":
            lib = glob.glob(
                os.path.join(
                    self.model_paths.tmp_dir,
                    "build",
                    "out",
                    "library",
                    "Debug",
                    self.model_paths.module_name + ".*",
                )
            )
        else:
            lib = glob.glob(
                os.path.join(
                    self.model_paths.tmp_dir,
                    "build",
                    "out",
                    "library",
                    self.model_paths.module_name + ".*",
                )
            )
        return len(lib) == 0

    def _get_simulink_deps(self):
        """
        Generates a list of all simulink dependencies and their paths
        """
        files = glob.glob(self.model_paths.simulink_native + "/**/*.h", recursive=True)

        self.simulink_deps = {os.path.basename(f).split(".")[0] for f in files}
        self.simulink_deps_path = files

        simulink_deps = glob.glob(self.model_paths.simulink_native + "/**/*.c", recursive=True)
        rt_main = None
        for file in simulink_deps:
            if os.path.basename(file) in ["rt_main.c", "classic_main.c"]:
                rt_main = file
                break
        if rt_main is not None:
            simulink_deps.remove(rt_main)

        for file in self.simulink_deps:
            if file == "rtw_matlogging":
                self.matlogging = True
                break

        self.simulink_deps_src = simulink_deps

    def _gen_custom_srcs(self):
        """
        Moves all custom mixin source files to the temporary directory and makes appropriate replacements
        in the source files
        """
        shutil.rmtree(
            os.path.join(self.model_paths.tmp_dir, "c_files"),
            ignore_errors=True,
        )
        shutil.copytree(
            os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "c_files")),
            os.path.join(self.model_paths.tmp_dir, "c_files"),
        )
        self.custom_includes = os.path.join(self.model_paths.tmp_dir, "c_files", "include")
        self.custom_sources = os.path.join(self.model_paths.tmp_dir, "c_files", "src")

        replacements = {
            "<<ROOT_MODEL>>": self.model_paths.root_model_name + ".h",
            "<<ROOT_MODEL_PRIVATE>>": self.model_paths.root_model_name + "_private.h",
            "<<MODEL_INTERFACE_C>>": self.model_paths.module_name,
            "<<ROOT_MODEL_NAME>>": sanitize_model_name(self.model_paths.root_model_name),
            "<<DATA_TYPE>>": self.gather_types(),
            "<<ALL_DTYPES>>": self.get_type_names(),
            "<<RTW_MATLOGGING>>": '#include "rtw_matlogging.h"' if self.matlogging else "",
        }
        self._replace_macros(os.path.join(self.custom_includes, "model_utils.hpp"), replacements)
        self._replace_macros(
            os.path.join(self.custom_includes, "model_interface.hpp"), replacements
        )
        self._replace_macros(os.path.join(self.custom_sources, "bindings.cpp"), replacements)
        self._replace_macros(os.path.join(self.custom_includes, "containers.hpp"), replacements)

        defines = os.path.join(self.model_paths.root_model_path, "defines.txt")
        if os.path.exists(defines):
            with open(defines, "r", encoding="utf-8") as f:
                self.defines = [line.strip() for line in f.readlines()]
        else:
            self.defines = infer_defines(self.model_paths)

    def _build(self):
        """
        Cals cmake to configure and build the extension. Writes errors to the current working directory
        in a log file.
        """
        build_dir = os.path.join(self.model_paths.tmp_dir, "build")

        with Popen(
            [
                os.path.join(cmake.CMAKE_BIN_DIR, "cmake"),
                "-S",
                self.model_paths.tmp_dir,
                "-DCMAKE_BUILD_TYPE=Release",
                '-G',
                self.generator,
                "-B",
                build_dir,
            ],
            stdout=PIPE,
            stderr=PIPE,
        ) as p:
            (output1, err1) = p.communicate()
            build = p.wait()

        if build != 0:
            now = datetime.now()
            err_file = os.path.join(
                os.getcwd(),
                now.strftime("%Y-%m-%d_%H-%M-%S_PySimlink_Generation_Error.log"),
            )
            with open(err_file, "w", encoding="utf-8") as f:
                f.write(output1.decode() if output1 else "")
                f.write(err1.decode() if err1 else "")
            raise GenerationError(
                err_file, os.path.join(self.model_paths.tmp_dir, "CMakeLists.txt")
            )

        with Popen(
            [os.path.join(cmake.CMAKE_BIN_DIR, "cmake"), "--build", build_dir],
            stdout=PIPE,
            stderr=PIPE,
        ) as p:
            (output2, err2) = p.communicate()
            make = p.wait()

        if make != 0:
            now = datetime.now()
            err_file = os.path.join(
                os.getcwd(),
                now.strftime("%Y-%m-%d_%H-%M-%S_PySimlink_Build_Error.log"),
            )
            with open(err_file, "w", encoding="utf-8") as f:
                f.write(output2.decode() if output2 else "")
                f.write(err2.decode() if err2 else "")

            raise BuildError(err_file, os.path.join(self.model_paths.tmp_dir, "CMakeLists.txt"))

    @staticmethod
    def _replace_macros(path: str, replacements: "dict[str, str]"):
        """
        Replaces strings in a file

        Args:
            path: path to file to replace strings in
            replacements: dictionary of replacements

        """
        with open(path, "r", encoding="utf-8") as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            for key, val in replacements.items():
                lines[i] = lines[i].replace(str(key), str(val))

        with open(path, "w", encoding="utf-8") as f:
            f.writelines(lines)

    def _read_types_single_file(self, lines):
        define_re = re.compile(r"^#define DEFINED_TYPEDEF")
        endif = re.compile(r"^#endif")
        pairs = []
        cur = None
        for i, line in enumerate(lines):
            test1 = re.search(define_re, line)
            test2 = re.search(endif, line)
            if test1 is not None:
                if cur is not None:
                    warnings.warn("types file malformed")
                else:
                    cur = i
                continue
            if test2 is not None:
                if cur is not None:
                    pairs.append((cur, i))
                    cur = None

        for pair in pairs:
            new_struct = parse_struct(lines[pair[0] + 2 : pair[1] - 1])
            for struct in self.types:
                ## Prevent duplicate types
                if struct.name == new_struct.name:
                    break
            else:
                self.types.append(new_struct)

    def _gen_types(self):
        ret = []
        for type in self.types:
            ret += [f'    py::class_<{type.name}>(m, "{type.name}", py::module_local())']
            for field in type.fields:
                ret += [f'            .def_readonly("{field.name}", &{type.name}::{field.name})']
            ret[-1] += ";"
            ret.append("")

        ret += [
            f'    py::class_<PYSIMLINK::all_dtypes>(m, "{sanitize_model_name(self.model_paths.root_model_name)}_all_dtypes", py::module_local())'
        ]
        for type in self.types:
            ret += [
                f'            .def_readonly("{type.name}", &PYSIMLINK::all_dtypes::{type.name}_obj)'
            ]
        ret[-1] += ";"
        ret.append("")

        return "\n".join(ret)

    def get_type_names(self):
        ret = []
        for struct in self.types:
            ret.append(f"{struct.name} {struct.name}_obj;")

        return "\n        ".join(ret)
